/*
 * SSE version.
 */

#include <stdio.h>
#include <math.h>
#include <emmintrin.h>
#include "../config.h"
#include "../utils.h"

float *compute_jacobi(int n, float init_value, float threshold, borders b, int *iterations) {
  float *x;
  float *new_x;
  float *tmp_x;
  float max_diff, new_value;
  int i, j, k;
  int nb = n + 2; // n plus the border
  int n_mult = (n / 4) * 4;
  __m128 new_value_vec, tmp_vec, diff_vec;
  float diffs[] = {0.0, 0.0, 0.0, 0.0};
  /* Initialize boundary regions */
  x = create_sa_matrix(nb, nb);
  new_x = create_sa_matrix(nb, nb);
  for (i = 0; i < nb; i++) {
    x[IDX(nb, 0, i)] = b.north;
    x[IDX(nb, n + 1, i)] = b.south;
    x[IDX(nb, i, 0)] = b.west;
    x[IDX(nb, i, n + 1)] = b.east;
    new_x[IDX(nb, 0, i)] = b.north;
    new_x[IDX(nb, n + 1, i)] = b.south;
    new_x[IDX(nb, i, 0)] = b.west;
    new_x[IDX(nb, i, n + 1)] = b.east;
  }
  /* Initialize the rest of the matrix */
  for (i = 1; i <= n; i++) {
    for (j = 1; j <= n; j++) {
      x[IDX(nb, i, j)] = init_value;
    }
  }
  /* Iterative refinement of x until values converge */
  *iterations = 0;
  do {
    max_diff = 0;
    for (i = 1; i <= n; i++) {
      for (j = 1; j <= n_mult; j += 4) {
        new_value_vec = _mm_loadu_ps(&x[IDX(nb, i - 1, j)]);
        tmp_vec = _mm_loadu_ps(&x[IDX(nb, i + 1, j)]);
        new_value_vec = _mm_add_ps(new_value_vec, tmp_vec);
        tmp_vec = _mm_loadu_ps(&x[IDX(nb, i, j - 1)]);
        new_value_vec = _mm_add_ps(new_value_vec, tmp_vec);
        tmp_vec = _mm_loadu_ps(&x[IDX(nb, i, j + 1)]);
        new_value_vec = _mm_add_ps(new_value_vec, tmp_vec);
        tmp_vec = _mm_set1_ps(0.25);
        new_value_vec = _mm_mul_ps(new_value_vec, tmp_vec);
        _mm_storeu_ps(&new_x[IDX(nb, i, j)], new_value_vec);

        diff_vec = _mm_loadu_ps(&new_x[IDX(nb, i, j)]);
        tmp_vec = _mm_loadu_ps(&x[IDX(nb, i, j)]);
        diff_vec = _mm_sub_ps(diff_vec, tmp_vec);
        _mm_storeu_ps(diffs, diff_vec);
        for (k = 0; k < 4; k++) {
          max_diff = fmaxf(max_diff, fabs(diffs[k]));
        }
      }
      for (j = n_mult; j <= n; j++) {
        new_value = 0.25 * (x[IDX(nb, i - 1, j)] + x[IDX(nb, i, j + 1)] + x[IDX(nb, i + 1, j)] + x[IDX(nb, i, j - 1)]);
        max_diff = fmaxf(max_diff, fabs(new_value - x[IDX(nb, i, j)]));
        new_x[IDX(nb, i, j)] = new_value;
      }
    }
    tmp_x = new_x;
    new_x = x;
    x = tmp_x;
    (*iterations)++;
  } while (max_diff > threshold);
  return x;
}
